from .bayesGM import CausalBGM
from .util import parse_file, save_data
import argparse
from BayesGM import __version__

def main(args=None):
    # General parameters
    parser = argparse.ArgumentParser('CausalBGM',
                                     description=f'CausalBGM: An AI-powered Bayesian generative modeling approach for causal inference in observational studies - v{__version__}')
    parser.add_argument('-o', '--output_dir', type=str, required=True,
                        help="Output directory")
    parser.add_argument('-i', '--input', type=str, required=True,
                        help="Input data file must be in csv or txt or npz format")
    parser.add_argument('-t', '--delimiter', type=str, default='\t',
                        help="Delimiter for txt or csv files (default: tab '\\t').")
    parser.add_argument('-d', '--dataset', type=str,default='Mydata',
                        help="Dataset name")
    parser.add_argument('-F', '--save_format', type=str,default='txt',
                        help="Saving format (default: txt)")
    parser.add_argument('-save_model', default=False, action=argparse.BooleanOptionalAction,
                        help="whether to save model.")
    parser.add_argument('-save_res', default=True, action=argparse.BooleanOptionalAction,
                        help="whether to save intermediate results.")
    parser.add_argument('-B', '--binary_treatment', default=True, action=argparse.BooleanOptionalAction,
                        help="whether use binary treatment setting.")
    parser.add_argument('--use_egm_init', default=True, action=argparse.BooleanOptionalAction,
                        help="whether use EGM initialization.")
    parser.add_argument('--use_bnn', default=True, action=argparse.BooleanOptionalAction,
                        help="whether use Bayesian neural nets.")

    # Parameters for iterative updating algorithm
    parser.add_argument('-Z', '--z_dims', type=int, nargs='+', default=[3,3,6,6],
                        help='Latent dimensions of Z (default: [3, 3, 6, 6]).')
    parser.add_argument('--lr_theta', type=float, default=0.0001,
                        help="Learning rate for updating model parameters (default: 0.0001).")
    parser.add_argument('--lr_z', type=float, default=0.0001,
                        help="Learning rate for updating latent variables (default: 0.0001).")
    parser.add_argument('--x_min', type=float, default=0.,
                        help="Lower bound for treatment interval (default: 0.0).")
    parser.add_argument('--x_max', type=float, default=3.,
                        help="Upper bound for treatment interval (default: 3.0).")
    parser.add_argument('--x_values', type=float, nargs='+',
                        help="List of treatment values to be predicted. Provide space-separated values. Example: --x_values 0.5 1.0 1.5")
    parser.add_argument('--g_units', type=int, nargs='+', default=[64,64,64,64,64],
                        help='Number of units for covariates generative model (default: [64,64,64,64,64]).')
    parser.add_argument('--f_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for outcome generative model (default: [64,32,8]).')
    parser.add_argument('--h_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for treatment generative model (default: [64,32,8]).')

    # Parameters for EGM initialization
    parser.add_argument('--kl_weight', type=float, default=0.0001,
                        help="coefficient for KL divergence term in BNNs (default: 0.0001).")
    parser.add_argument('--lr', type=float, default=0.0001,
                        help="Learning rate for EGM initialization (default: 0.0002).")
    parser.add_argument('--g_d_freq', type=int, default=5,
                        help="Frequency for updating discriminators and generators (default: 5).")
    parser.add_argument('--e_units', type=int, nargs='+', default=[64,64,64,64,64],
                        help='Number of units for encoder network (default: [64,64,64,64,64]).')
    parser.add_argument('--dz_units', type=int, nargs='+', default=[64,32,8],
                        help='Number of units for discriminator network in latent space (default: [64,32,8]).')
    parser.add_argument('--use-z-rec', default=True, action=argparse.BooleanOptionalAction,
                        help="Use the reconstruction for latent features (default: True).")

    # Parameters for fitting && predicting
    parser.add_argument('-N','--n_iter', type=int, default=30000,
                        help="Number of iterations (default: 30000).")
    parser.add_argument('--startoff', type=int, default=0,
                        help="Iteration for starting evaluation (default: 0).")
    parser.add_argument('--batches_per_eval', type=int, default=500,
                        help="Number of iterations per evaluation (default: 500).")
    parser.add_argument('-E', '--epochs', type=int, default=100,
                        help="Number of epochs in iterative updating algorithm (default: 100).")
    parser.add_argument('-M', '--n_mcmc', type=int, default=3000,
                        help="MCMC sample size (default: 3000).")
    parser.add_argument('--epochs_per_eval', type=int, default=10,
                        help="Number of epochs per evaluation (default: 10).")
    parser.add_argument('--alpha', type=float, default=0.01,
                        help="Significant level (default: 0.01).")

    #Random seed control
    parser.add_argument('--seed', type=int, default=123,
                        help="Random seed for reproduction (default: 123).")
    args = parser.parse_args()
    params = vars(args)
    data = parse_file(args.input, sep=params['delimiter'])
    params['v_dim'] = data[-1].shape[1]

    # Instantiate a CausalBGM model
    model = CausalBGM(params=params, random_seed=None)

    # Perform Encoding Generative Modeling (EGM) initialization
    # n_iter=30000: Number of iterations for the initialization process
    # batches_per_eval=500: Frequency of evaluations (e.g., every 500 batches)
    # verbose=1: Controls verbosity level, showing progress and evaluation metrics
    model.egm_init(data=data, n_iter=params['n_iter'], batches_per_eval=params['batches_per_eval'], verbose=1)

    # Train the CausalBGM model with an iterative updating algorithm
    # epochs=100: Total number of training epochs
    # epochs_per_eval=10: Frequency of evaluation during training (e.g., every 10 epochs)
    model.fit(data=data, epochs=params['epochs'], epochs_per_eval=params['epochs_per_eval'])

    # Make predictions using the trained CausalBGM model
    # alpha=0.01: Significance level for the posterior intervals
    # n_mcmc=3000: Number of MCMC posterior samples for inference
    # x_values: treatment values to be predicted for ADRF
    # q_sd=1.0: Standard deviation for the proposal distribution in Metropolis-Hastings sampling,q_sd=-1 enables adaptive S.D.
    # Returns:
    #   causal_pre: Estimated causal effects (ADRF for continuous treatment) with shape (len(x_values),)
    #   pos_intervals: Posterior intervals for the estimated causal effects with shape (len(x_values), 2)
    if params['binary_treatment']:
        causal_pre, pos_intervals = model.predict(data=data, alpha=params['alpha'], n_mcmc=params['n_mcmc'], q_sd=1.0)
    else:
        causal_pre, pos_intervals = model.predict(data=data, alpha=params['alpha'], n_mcmc=params['n_mcmc'], x_values=params['x_values'], q_sd=1.0)

    # Save results to 'save_dir'
    save_data('{}/causal_effect_point_estimate.{}'.format(model.save_dir, params['save_format']), causal_pre)
    save_data('{}/causal_effect_poterior_interval.{}'.format(model.save_dir, params['save_format']), pos_intervals)

if __name__ == "__main__":
   main()



import numpy as np
import math
import os
import sys
import pandas as pd
from sklearn.preprocessing import MinMaxScaler,MaxAbsScaler,StandardScaler
from sklearn.model_selection import train_test_split
from sklearn import datasets
from scipy.sparse import diags
from scipy.stats import norm

def Dataset_selector(name):
    if name == 'Semi_acic':
        return Semi_acic_sampler
    elif name=='Semi_ihdp':
        return Semi_ihdp_sampler
    elif name=='Sim_Hirano_Imbens':
        return Sim_Hirano_Imbens_sampler
    elif name=='Sim_Sun':
        return Sim_Sun_sampler
    elif name=='Sim_Colangelo':
        return Sim_Colangelo_sampler
    elif name=='Semi_Twins':
        return Semi_Twins_sampler
    else:
        print('Cannot find the example data sampler: %s!'%name)
        sys.exit()

class Base_sampler(object):
    def __init__(self, x, y, v, batch_size=32, normalize=False, random_seed=123):
        assert len(x)==len(y)==len(v)
        np.random.seed(random_seed)
        self.data_x = np.array(x, dtype='float32')
        self.data_y = np.array(y, dtype='float32')
        self.data_v = np.array(v, dtype='float32')
        if len(self.data_x.shape) == 1:
            self.data_x = self.data_x.reshape(-1,1)
        if len(self.data_y.shape) == 1:
            self.data_y = self.data_y.reshape(-1,1)
        self.batch_size = batch_size
        if normalize:
            self.data_v = StandardScaler().fit_transform(self.data_v)
            #self.data_v = MinMaxScaler().fit_transform(self.data_v)
        self.sample_size = len(x)
        self.full_index = np.arange(self.sample_size)
        np.random.shuffle(self.full_index)
        self.idx_gen = self.create_idx_generator(sample_size=self.sample_size)
        
    def create_idx_generator(self, sample_size, random_seed=123):
        while True:
            for step in range(math.ceil(sample_size/self.batch_size)):
                if (step+1)*self.batch_size <= sample_size:
                    yield self.full_index[step*self.batch_size:(step+1)*self.batch_size]
                else:
                    yield np.hstack([self.full_index[step*self.batch_size:],
                                    self.full_index[:((step+1)*self.batch_size-sample_size)]])
                    np.random.shuffle(self.full_index)

    def next_batch(self):
        indx = next(self.idx_gen)
        return self.data_x[indx,:], self.data_y[indx,:], self.data_v[indx, :]
    
    def load_all(self):
        return self.data_x, self.data_y, self.data_v

class Semi_acic_sampler(Base_sampler):
    """ACIC 2018 competition dataset (binary treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    path
        Str object denoting the path to the original dataset.
    ufid
        Str object denoting the unique id of a specific semi-synthetic setting.
    Examples
    --------
    >>> from CausalEGM import Semi_acic_sampler
    >>> import numpy as np
    >>> x = np.random.normal(size=(2000,))
    >>> y = np.random.normal(size=(2000,))
    >>> v = np.random.normal(size=(2000,100))
    >>> ds = Semi_acic_sampler(path='../data/ACIC_2018',ufid='d5bd8e4814904c58a79d7cdcd7c2a1bb')
    """
    def __init__(self, batch_size=32, path='../data/ACIC_2018', 
                ufid='d5bd8e4814904c58a79d7cdcd7c2a1bb'):
        self.df_covariants = pd.read_csv('%s/x.csv'%path, index_col='sample_id',header=0, sep=',')
        self.df_sim = pd.read_csv('%s/scaling/factuals/%s.csv'%(path, ufid),index_col='sample_id',header=0, sep=',')
        dataset = self.df_covariants.join(self.df_sim, how='inner')
        x = dataset['z'].values.reshape(-1,1)
        y = dataset['y'].values.reshape(-1,1)
        v = dataset.values[:,:-2]
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Hirano_Imbens_sampler(Base_sampler):
    """Hirano Imbens simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Hirano_Imbens_sampler
    >>> ds = Sim_Hirano_Imbens_sampler(batch_size=32, N=20000, v_dim=200, seed=0)
    """
    def __init__(self, batch_size=32, N=20000, v_dim=200, seed=0):
        np.random.seed(seed)
        v = np.random.exponential(scale=1.0, size=(N, v_dim))
        rate = v[:,0] + v[:,1]
        scale = 1/rate
        x = np.random.exponential(scale=scale)
        y = np.random.normal(x + (v[:,0] + v[:,2]) * np.exp(-x * (v[:,0] + v[:,2])) , 1)
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Sun_sampler(Base_sampler):
    """Sun simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Sun_sampler
    >>> ds = Sim_Sun_sampler(batch_size=32, N=20000, v_dim=200, seed=0)
    """
    def __init__(self, batch_size=32, N=20000, v_dim=200, seed=0):
        np.random.seed(seed)
        v = np.random.normal(0, 1, size=(N, v_dim))        
        x = np.random.normal(-2*(np.sin(2*v[:,0]))+ ((v[:,1])**2 - 1/3) + (v[:,2]-1/2) + np.cos(v[:,3]), 1)
        y = np.random.normal(((v[:,0] - 1/2)+ np.cos(v[:,1]) + (v[:,4])**2 + (v[:,5])) + x, 1)       
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Sim_Colangelo_sampler(Base_sampler):
    """Colangelo simulation dataset (continuous treatment) sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    N
        Sample size. Default: ``20000``.
    v_dim
        Int object denoting the dimension for covariates. Default: ``200``.
    seed
        Int object denoting the random seed. Default: ``0``.
    Examples
    --------
    >>> from CausalEGM import Sim_Colangelo_sampler
    >>> ds = Sim_Colangelo_sampler(batch_size=32, N=20000, v_dim=100, seed=0)
    """
    def __init__(self, batch_size=32, N=20000, v_dim=100, seed=0,
                rho=0.5, offset = [-1,0,1], d=1, a=3, b=0.75):
        np.random.seed(seed)
        k = np.array([rho*np.ones(v_dim-1),np.ones(v_dim),rho*np.ones(v_dim-1)],dtype=object)
        sigma = diags(k,offset).toarray()
        theta = np.array([(1/(l**2)) for l in list(range(1,(v_dim+1)))])
        epsilon = np.random.normal(0,1,N)
        nu = np.random.normal(0,1,N)
        v = np.random.multivariate_normal(np.zeros(v_dim),sigma,size=[N,])
        x = d*norm.cdf((a*v@theta)) + b*nu - 0.5
        y = 1.2*x + (x**3) + (x*v[:,0]) + 1.2*(v@theta) + epsilon
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Semi_Twins_sampler(Base_sampler):
    """Twins semi synthetic  dataset sampler (inherited from Base_sampler).

    Parameters
    ----------
    batch_size
        Int object denoting the batch size for mini-batch training. Default: ``32``.
    seed
        Int object denoting the random seed. Default: ``0``.
    path
        Str obejct denoting the path to the original data.
    Examples
    --------
    >>> from CausalEGM import Semi_Twins_sampler
    >>> ds = Semi_Twins_sampler(batch_size=32, path='../data/Twins')
    """
    def __init__(self, batch_size=32, seed=0,
                path='../data/Twins'):
        covariate_df = pd.read_csv('%s/twin_pairs_X_3years_samesex.csv'%path).iloc[:,2:].drop(['infant_id_0', 'infant_id_1'], axis=1)
        treatment_df_ = pd.read_csv('%s/twin_pairs_T_3years_samesex.csv'%path).iloc[:,1:]
        outcome_df = pd.read_csv('%s/twin_pairs_Y_3years_samesex.csv'%path).iloc[:,1:]
        #### discard NAN values
        rows_with_nan = [index for index, row in covariate_df.iterrows() if row.isnull().any()]
        covariate_df = covariate_df.drop(rows_with_nan)
        treatment_df_ = treatment_df_.drop(rows_with_nan)
        outcome_df = outcome_df.drop(rows_with_nan)
        #### select those below 2kg:
        rows_less2kg = [index for index, row in treatment_df_.iterrows() if (row['dbirwt_1']>=2000)]
        covariate_df = covariate_df.drop(rows_less2kg)
        treatment_df_ = treatment_df_.drop(rows_less2kg)
        outcome_df = outcome_df.drop(rows_less2kg)

        x = np.concatenate([treatment_df_.values[:,0], treatment_df_.values[:,1]])/1000
        v =  np.concatenate([covariate_df.values, covariate_df.values])
        np.random.seed(seed)
        eps = np.random.normal(0, 0.25, size=(v.shape[0],))
        gamma = np.random.normal(0, 0.025, size=(v.shape[1],))
        y = -2 * 1/(1 + np.exp(-3 * x)) + np.dot(v, gamma) + eps
        self.auxiliary_constant =  np.mean(np.dot(v, gamma))
        x = x.reshape(-1,1)
        y = y.reshape(-1,1)
        super().__init__(x,y,v,batch_size=batch_size,normalize=True)

class Gaussian_sampler(object):
    def __init__(self, mean, sd=1, N=20000):
        self.total_size = N
        self.mean = mean
        self.sd = sd
        np.random.seed(1024)
        self.X = np.random.normal(self.mean, self.sd, (self.total_size,len(self.mean)))
        self.X = self.X.astype('float32')
        self.Y = None

    def train(self, batch_size, label = False):
        indx = np.random.randint(low = 0, high = self.total_size, size = batch_size)
        return self.X[indx, :]

    def get_batch(self,batch_size):
        return np.random.normal(self.mean, self.sd, (batch_size,len(self.mean))).astype('float32')

    def load_all(self):
        return self.X, self.Y

def save_data(fname, data):

    """Save the data to the specified path."""
    if fname[-3:] == 'npy':
        np.save(fname, data)
    elif fname[-3:] == 'txt' or 'csv':
        np.savetxt(fname, data, fmt='%.6f')
    else:
        print('Wrong saving format, please specify either .npy, .txt, or .csv')
        sys.exit()

def parse_file(path, sep='\t', header = 0, normalize=True):
    """
    Parse an input file and extract features (x, y, and v) for model training or evaluation.

    Parameters:
    -----------
    path : str
        Path to the input file. The file can be in .npz, .csv, or .txt format.
    sep : str, optional (default: '\t')
        Separator used in .csv or .txt files. Defaults to tab-delimited format ('\t').
    header : int or None, optional (default: 0)
        Row number to use as column names in .csv files. Default is 0 (the first row). 
        Use `None` if the file does not have a header.
    normalize : bool, optional (default: True)
        If True, the features in `v` will be normalized using `StandardScaler`.

    Returns:
    --------
    data_x : np.ndarray
        The treatment variable(s) extracted from the file, reshaped to (-1, 1).
    data_y : np.ndarray
        The outcome variable(s) extracted from the file, reshaped to (-1, 1).
    data_v : np.ndarray
        Covariates extracted from the file. Normalized if `normalize=True`.

    Notes:
    ------
    - Supported file formats:
        - `.npz`: Numpy compressed files with keys `x`, `y`, and `v`.
        - `.csv`: Comma-separated value files with treatment, outcome, and covariates as columns.
        - `.txt`: Tab- or other character-delimited text files with similar structure to .csv.
    - The input file must exist at the specified `path`.
    - The first column is assumed to be the treatment variable (`x`).
    - The second column is assumed to be the outcome variable (`y`).
    - Remaining columns are assumed to be covariates (`v`).

    Example:
    --------
    # Example for .csv input
    data_x, data_y, data_v = parse_file("data.csv", sep=',', header=0, normalize=True)
    
    # Example for .npz input
    data_x, data_y, data_v = parse_file("data.npz", normalize=False)
    """
    assert os.path.exists(path)
    if path[-3:] == 'npz':
        data = np.load(path)
        data_x, data_y, data_v = data['x'],data['y'],data['v']
    elif  path[-3:] == 'csv':
        data = pd.read_csv(path, header=0, sep=sep).values
        data_x = data[:,0].reshape(-1, 1).astype('float32')
        data_y = data[:,1].reshape(-1, 1).astype('float32')
        data_v = data[:,2:].astype('float32')
    elif path[-3:] == 'txt':
        data = np.loadtxt(path,delimiter=sep)
        data_x = data[:,0].reshape(-1, 1).astype('float32')
        data_y = data[:,1].reshape(-1, 1).astype('float32')
        data_v = data[:,2:].astype('float32')
    else:
        print('File format not recognized, please use .npz, .csv or .txt as input.')
        sys.exit()
    if normalize:
        data_v = StandardScaler().fit_transform(data_v)
    return data_x, data_y, data_v



import setuptools

setuptools.setup(
    name="CausalBGM", 
    version="0.1.4",
    author="Qiao Liu",
    author_email="liuqiao@stanford.edu",
    description="CausalBGM: An AI-powered Bayesian generative modeling approach for causal inference in observational studies",
    long_description="Causal inference in observational studies with high-dimensional covariates presents significant challenges. We introduce CausalBGM, an AI-powered Bayesian generative modeling approach that captures the causal relationship among covariates, treatment, and outcome variables. The core innovation of CausalBGM lies in its ability to estimate the individual treatment effect (ITE) by learning individual-specific distributions of a low-dimensional latent feature set (e.g., latent confounders) that drives changes in both treatment and outcome. This approach not only effectively mitigates confounding effects but also provides comprehensive uncertainty quantification, offering reliable and interpretable causal effect estimates at the individual level. CausalBGM adopts a Bayesian model and uses a novel iterative algorithm to update the model parameters and the posterior distribution of latent features until convergence. This framework leverages the power of AI to capture complex dependencies among variables while adhering to the Bayesian principles. Extensive experiments demonstrate that CausalBGM consistently outperforms state-of-the-art methods, particularly in scenarios with high-dimensional covariates and large-scale datasets. Its Bayesian foundation ensures statistical rigor, providing robust and well-calibrated posterior intervals. By addressing key limitations of existing methods, CausalBGM emerges as a robust and promising framework for advancing causal inference in modern applications in fields such as genomics, healthcare, and social sciences.",
    long_description_content_type="text/markdown",
    url="https://github.com/SUwonglab/CausalBGM",
    packages=setuptools.find_packages(),
    install_requires=[
   'numpy==1.24.2',
   'tensorflow==2.10.0',
   'tensorflow-probability==0.18.0',
   'pyyaml',
   'scikit-learn',
   'pandas',
   'python-dateutil'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.9',
    entry_points={
    'console_scripts': [
        'causalBGM = BayesGM.cli:main',
    ]},
)
